@@ -8,25 +8,24 @@
 
 from biosimulators_utils.combine.exec import exec_sedml_docs_in_archive
 from biosimulators_utils.config import get_config, Config  # noqa: F401
-from biosimulators_utils.log.data_model import CombineArchiveLog, TaskLog  # noqa: F401
+from biosimulators_utils.log.data_model import CombineArchiveLog, TaskLog, StandardOutputErrorCapturerLevel  # noqa: F401
 from biosimulators_utils.model_lang.rba.validation import validate_model
 from biosimulators_utils.report.data_model import ReportFormat, VariableResults, SedDocumentResults  # noqa: F401
 from biosimulators_utils.sedml import validation
 from biosimulators_utils.sedml.data_model import (Task, ModelLanguage, ModelAttributeChange,  # noqa: F401
                                                   SteadyStateSimulation, UniformTimeCourseSimulation,
                                                   Variable, Symbol)
-from biosimulators_utils.sedml.exec import exec_sed_doc
+from biosimulators_utils.sedml.exec import exec_sed_doc as base_exec_sed_doc
 from biosimulators_utils.simulator.utils import get_algorithm_substitution_policy
 from biosimulators_utils.utils.core import raise_errors_warnings
 from biosimulators_utils.viz.data_model import VizFormat  # noqa: F401
 from biosimulators_utils.warnings import warn, BioSimulatorsWarning
 from kisao.data_model import AlgorithmSubstitutionPolicy, ALGORITHM_SUBSTITUTION_POLICY_LEVELS
 from kisao.utils import get_preferred_substitute_algorithm_by_ids
-import functools
 import numpy
 import rba
 
-__all__ = ['exec_sedml_docs_in_combine_archive', 'exec_sed_task']
+__all__ = ['exec_sedml_docs_in_combine_archive', 'exec_sed_doc', 'exec_sed_task', 'preprocess_sed_task', 'modify_model']
 
 
 def exec_sedml_docs_in_combine_archive(archive_filename, out_dir, config=None):
@@ -49,20 +48,65 @@ def exec_sedml_docs_in_combine_archive(archive_filename, out_dir, config=None):
             * :obj:`SedDocumentResults`: results
             * :obj:`CombineArchiveLog`: log
     """
-    sed_doc_executer = functools.partial(exec_sed_doc, exec_sed_task)
-    return exec_sedml_docs_in_archive(sed_doc_executer, archive_filename, out_dir,
+    return exec_sedml_docs_in_archive(exec_sed_doc, archive_filename, out_dir,
                                       apply_xml_model_changes=False,
                                       config=config)
 
 
-def exec_sed_task(task, variables, log=None, config=None):
+def exec_sed_doc(doc, working_dir, base_out_path, rel_out_path=None,
+                 apply_xml_model_changes=False,
+                 log=None, indent=0, pretty_print_modified_xml_models=False,
+                 log_level=StandardOutputErrorCapturerLevel.c, config=None):
+    """ Execute the tasks specified in a SED document and generate the specified outputs
+
+    Args:
+        doc (:obj:`SedDocument` or :obj:`str`): SED document or a path to SED-ML file which defines a SED document
+        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)
+
+        base_out_path (:obj:`str`): path to store the outputs
+
+            * CSV: directory in which to save outputs to files
+              ``{base_out_path}/{rel_out_path}/{report.id}.csv``
+            * HDF5: directory in which to save a single HDF5 file (``{base_out_path}/reports.h5``),
+              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file
+
+        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the outputs
+        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
+            calling :obj:`task_executer`.
+        log (:obj:`SedDocumentLog`, optional): log of the document
+        indent (:obj:`int`, optional): degree to indent status messages
+        pretty_print_modified_xml_models (:obj:`bool`, optional): if :obj:`True`, pretty print modified XML models
+        log_level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which to log output
+        config (:obj:`Config`, optional): BioSimulators common configuration
+        simulator_config (:obj:`SimulatorConfig`, optional): tellurium configuration
+
+    Returns:
+        :obj:`tuple`:
+
+            * :obj:`ReportResults`: results of each report
+            * :obj:`SedDocumentLog`: log of the document
+    """
+    return base_exec_sed_doc(exec_sed_task, doc, working_dir, base_out_path,
+                             rel_out_path=rel_out_path,
+                             apply_xml_model_changes=apply_xml_model_changes,
+                             log=log,
+                             indent=indent,
+                             pretty_print_modified_xml_models=pretty_print_modified_xml_models,
+                             log_level=log_level,
+                             config=config)
+
+
+def exec_sed_task(task, variables, preprocessed_task=None, log=None, config=None):
     """ Execute a task and save its results
 
     Args:
-       task (:obj:`Task`): task
-       variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
-       log (:obj:`TaskLog`, optional): log for the task
-       config (:obj:`Config`, optional): BioSimulators common configuration
+        task (:obj:`Task`): task
+        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
+        preprocessed_task (:obj:`dict`, optional): preprocessed information about the task, including possible
+            model changes and variables. This can be used to avoid repeatedly executing the same initialization
+            for repeated calls to this method.
+        log (:obj:`TaskLog`, optional): log for the task
+        config (:obj:`Config`, optional): BioSimulators common configuration
 
     Returns:
         :obj:`tuple`:
@@ -81,6 +125,62 @@ def exec_sed_task(task, variables, log=None, config=None):
     if config.LOG and not log:
         log = TaskLog()
 
+    if preprocessed_task is None:
+        preprocessed_task = preprocess_sed_task(task, variables, config=config)
+
+    # validate task
+    model = task.model
+
+    # read model
+    rba_model = preprocessed_task['model']
+
+    # modify model
+    modify_model(rba_model, model.changes, preprocessed_task)
+
+    # instantiate simulation
+    rba_results = rba_model.solve()
+
+    # transform simulation results
+    variable_results = VariableResults()
+    for variable in variables:
+        variable_type, _, rba_id = variable.target.partition('.')
+
+        if variable_type == 'objective':
+            variable_results[variable.id] = numpy.array(rba_results.mu_opt)
+
+        elif variable_type == 'variables':
+            variable_results[variable.id] = numpy.array(rba_results.variables[rba_id])
+
+        else:
+            variable_results[variable.id] = numpy.array(rba_results.dual_values[rba_id])
+
+    # log action
+    if config.LOG:
+        log.algorithm = preprocessed_task['algorithm_kisao_id']
+
+        log.simulator_details = {}
+        log.simulator_details['method'] = '{}.{}.{}'.format(rba_model.__module__, rba_model.__class__.__name__, rba_model.solve.__name__)
+        log.simulator_details['lpSolver'] = rba_results._solver.lp_solver.name
+
+    ############################
+    # return the result of each variable and log
+    return variable_results, log
+
+
+def preprocess_sed_task(task, variables, config=None):
+    """ Preprocess a SED task, including its possible model changes and variables. This is useful for avoiding
+    repeatedly initializing tasks on repeated calls of :obj:`exec_sed_task`.
+
+    Args:
+        task (:obj:`Task`): task
+        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
+        config (:obj:`Config`, optional): BioSimulators common configuration
+
+    Returns:
+        :obj:`object`: preprocessed information about the task
+    """
+    config = config or get_config()
+
     # validate task
     model = task.model
     sim = task.simulation
@@ -108,6 +208,22 @@ def exec_sed_task(task, variables, log=None, config=None):
                           error_summary='Model `{}` is invalid.'.format(model.id),
                           warning_summary='Model `{}` may be invalid.'.format(model.id))
 
+    # validate changes
+    model_target_parameter_map = {}
+    for function in rba_model.parameters.functions:
+        for parameter in function.parameters:
+            target = 'parameters.functions.{}.parameters.{}'.format(function.id, parameter.id)
+            model_target_parameter_map[target] = parameter
+
+    invalid_changes = []
+    for i_change, change in enumerate(model.changes):
+        parameter = model_target_parameter_map.get(change.target, None)
+        if parameter is None:
+            invalid_changes.append('{}: {}'.format(i_change + 1, change.target))
+
+    if invalid_changes:
+        raise ValueError('The following changes are not valid:\n  - {}'.format('\n  - '.join(sorted(invalid_changes))))
+
     # validate variables
     constraint_matrix = rba.ConstraintMatrix(rba_model)
 
@@ -158,31 +274,24 @@ def exec_sed_task(task, variables, log=None, config=None):
             msg = "Algorithm changes were ignored because no algorithm parameters are supported."
             warn(msg, BioSimulatorsWarning)
 
-    # instantiate simulation
-    rba_results = rba_model.solve()
-
-    # transform simulation results
-    variable_results = VariableResults()
-    for variable in variables:
-        variable_type, _, rba_id = variable.target.partition('.')
-
-        if variable_type == 'objective':
-            variable_results[variable.id] = numpy.array(rba_results.mu_opt)
-
-        elif variable_type == 'variables':
-            variable_results[variable.id] = numpy.array(rba_results.variables[rba_id])
-
-        else:
-            variable_results[variable.id] = numpy.array(rba_results.dual_values[rba_id])
-
-    # log action
-    if config.LOG:
-        log.algorithm = exec_kisao_id
-
-        log.simulator_details = {}
-        log.simulator_details['method'] = '{}.{}.{}'.format(rba_model.__module__, rba_model.__class__.__name__, rba_model.solve.__name__)
-        log.simulator_details['lpSolver'] = rba_results._solver.lp_solver.name
-
     ############################
-    # return the result of each variable and log
-    return variable_results, log
+    # processed task
+    return {
+        'model': rba_model,
+        'model_target_parameter_map': model_target_parameter_map,
+        'algorithm_kisao_id': exec_kisao_id,
+    }
+
+
+def modify_model(model, changes, preprocessed_task):
+    """ Modify a model
+
+    Args:
+        model (:obj:`rba.model.RbaModel`): RBA model
+        changes (:obj:`list` of :obj:`ModelAttributeChange`): changes to apply to the model
+        preprocessed_task (:obj:`dict`): preprocessed informationa about the model
+    """
+    rba_model_target_parameter_map = preprocessed_task['model_target_parameter_map']
+    for change in changes:
+        parameter = rba_model_target_parameter_map[change.target]
+        parameter.value = float(change.new_value)
